from enum import Enum, EnumMeta


def _normalize_enum_value(value: str) -> str:
    return value.strip(" ").upper()


class _Parseable(EnumMeta):
    def __getitem__(self, name):
        if isinstance(name, str):
            name = _normalize_enum_value(name)
        return super().__getitem__(name)

    def __call__(cls, value, *args, **kw):
        if isinstance(value, str):
            value = _normalize_enum_value(value)
        return super().__call__(value, *args, **kw)

    def __contains__(self, child):
        return child in self.__members__


class ParseableEnum(Enum, metaclass=_Parseable):
    def __str__(self):
        return str(self.value)

    @classmethod
    def synonyms(cls):
        """Override to provide a dictionary of synonyms for the enum values."""
        return {}

    @classmethod
    def _missing_(cls, val):
        synonyms = cls.synonyms()
        if val in synonyms:
            return cls.__members__[synonyms[val]]
        return super()._missing_(val)


class ResourceType(ParseableEnum):
    ACCOUNT = "ACCOUNT"
    AGGREGATION_POLICY = "AGGREGATION POLICY"
    ALERT = "ALERT"
    API_INTEGRATION = "API INTEGRATION"
    APPLICATION_ROLE = "APPLICATION ROLE"
    AUTHENTICATION_POLICY = "AUTHENTICATION POLICY"
    CATALOG_INTEGRATION = "CATALOG INTEGRATION"
    CLASS = "CLASS"
    COLUMN = "COLUMN"
    COMPUTE_POOL = "COMPUTE POOL"
    DATABASE = "DATABASE"
    DATABASE_ROLE = "DATABASE ROLE"
    DIRECTORY_TABLE = "DIRECTORY TABLE"
    DYNAMIC_TABLE = "DYNAMIC TABLE"
    EVENT_TABLE = "EVENT TABLE"
    EXTERNAL_ACCESS_INTEGRATION = "EXTERNAL ACCESS INTEGRATION"
    EXTERNAL_FUNCTION = "EXTERNAL FUNCTION"
    FAILOVER_GROUP = "FAILOVER GROUP"
    FILE_FORMAT = "FILE FORMAT"
    FUNCTION = "FUNCTION"
    FUTURE_GRANT = "FUTURE GRANT"
    GIT_REPOSITORY = "GIT REPOSITORY"
    GRANT = "GRANT"
    GRANT_ON_ALL = "GRANT ON ALL"
    HYBRID_TABLE = "HYBRID TABLE"
    IMAGE_REPOSITORY = "IMAGE REPOSITORY"
    INTEGRATION = "INTEGRATION"
    MATERIALIZED_VIEW = "MATERIALIZED VIEW"
    NETWORK_POLICY = "NETWORK POLICY"
    NETWORK_RULE = "NETWORK RULE"
    NOTEBOOK = "NOTEBOOK"
    NOTIFICATION_INTEGRATION = "NOTIFICATION INTEGRATION"
    PACKAGES_POLICY = "PACKAGES POLICY"
    PASSWORD_POLICY = "PASSWORD POLICY"
    PIPE = "PIPE"
    PROCEDURE = "PROCEDURE"
    REPLICATION_GROUP = "REPLICATION GROUP"
    RESOURCE_MONITOR = "RESOURCE MONITOR"
    ROLE = "ROLE"
    ROLE_GRANT = "ROLE GRANT"
    SCHEMA = "SCHEMA"
    SECRET = "SECRET"
    SECURITY_INTEGRATION = "SECURITY INTEGRATION"
    SEQUENCE = "SEQUENCE"
    SERVICE = "SERVICE"
    SHARE = "SHARE"
    STAGE = "STAGE"
    STORAGE_INTEGRATION = "STORAGE INTEGRATION"
    STREAM = "STREAM"
    TABLE = "TABLE"
    TAG = "TAG"
    TAG_REFERENCE = "TAG REFERENCE"
    TASK = "TASK"
    USER = "USER"
    VIEW = "VIEW"
    WAREHOUSE = "WAREHOUSE"


class Scope(ParseableEnum):
    ORGANIZATION = "ORGANIZATION"
    ACCOUNT = "ACCOUNT"
    DATABASE = "DATABASE"
    SCHEMA = "SCHEMA"
    TABLE = "TABLE"


class AccountEdition(ParseableEnum):
    STANDARD = "STANDARD"
    ENTERPRISE = "ENTERPRISE"
    BUSINESS_CRITICAL = "BUSINESS-CRITICAL"


class DataType(ParseableEnum):
    NUMBER = "NUMBER"
    DECIMAL = "DECIMAL"
    NUMERIC = "NUMERIC"
    INT = "INT"
    INTEGER = "INTEGER"
    BIGINT = "BIGINT"
    SMALLINT = "SMALLINT"
    TINYINT = "TINYINT"
    BYTEINT = "BYTEINT"
    FLOAT = "FLOAT"
    FLOAT4 = "FLOAT4"
    FLOAT8 = "FLOAT8"
    DOUBLE = "DOUBLE"
    DOUBLE_PRECISION = "DOUBLE PRECISION"
    REAL = "REAL"
    VARCHAR = "VARCHAR"
    CHAR = "CHAR"
    CHARACTER = "CHARACTER"
    NCHAR = "NCHAR"
    STRING = "STRING"
    TEXT = "TEXT"
    NVARCHAR = "NVARCHAR"
    NVARCHAR2 = "NVARCHAR2"
    CHAR_VARYING = "CHAR VARYING"
    NCHAR_VARYING = "NCHAR VARYING"
    BINARY = "BINARY"
    VARBINARY = "VARBINARY"
    BOOLEAN = "BOOLEAN"
    DATE = "DATE"
    DATETIME = "DATETIME"
    TIME = "TIME"
    TIMESTAMP = "TIMESTAMP"
    TIMESTAMP_LTZ = "TIMESTAMP_LTZ"
    TIMESTAMP_NTZ = "TIMESTAMP_NTZ"
    TIMESTAMP_TZ = "TIMESTAMP_TZ"
    ARRAY = "ARRAY"
    OBJECT = "OBJECT"
    VARIANT = "VARIANT"
    GEOGRAPHY = "GEOGRAPHY"
    GEOMETRY = "GEOMETRY"


class Language(ParseableEnum):
    JAVA = "JAVA"
    JAVASCRIPT = "JAVASCRIPT"
    PYTHON = "PYTHON"
    SCALA = "SCALA"
    SQL = "SQL"


class NullHandling(ParseableEnum):
    CALLED_ON_NULL_INPUT = "CALLED ON NULL INPUT"
    RETURNS_NULL_ON_NULL_INPUT = "RETURNS NULL ON NULL INPUT"
    STRICT = "STRICT"


class Volatility(ParseableEnum):
    VOLATILE = "VOLATILE"
    IMMUTABLE = "IMMUTABLE"
    STABLE = "STABLE"


# https://docs.snowflake.com/developer-guide/stored-procedure/stored-procedures-rights
class ExecutionRights(ParseableEnum):
    CALLER = "CALLER"
    OWNER = "OWNER"


class SessionParameter(ParseableEnum):
    ABORT_DETACHED_QUERY = "ABORT_DETACHED_QUERY"
    AUTOCOMMIT = "AUTOCOMMIT"
    AUTOCOMMIT_API_SUPPORTED = "AUTOCOMMIT_API_SUPPORTED"
    BINARY_INPUT_FORMAT = "BINARY_INPUT_FORMAT"
    BINARY_OUTPUT_FORMAT = "BINARY_OUTPUT_FORMAT"
    CLIENT_ENABLE_CONSERVATIVE_MEMORY_USAGE = "CLIENT_ENABLE_CONSERVATIVE_MEMORY_USAGE"
    CLIENT_ENABLE_DEFAULT_OVERWRITE_IN_PUT = "CLIENT_ENABLE_DEFAULT_OVERWRITE_IN_PUT"
    CLIENT_ENABLE_LOG_INFO_STATEMENT_PARAMETERS = "CLIENT_ENABLE_LOG_INFO_STATEMENT_PARAMETERS"
    CLIENT_MEMORY_LIMIT = "CLIENT_MEMORY_LIMIT"
    CLIENT_METADATA_REQUEST_USE_CONNECTION_CTX = "CLIENT_METADATA_REQUEST_USE_CONNECTION_CTX"
    CLIENT_METADATA_USE_SESSION_DATABASE = "CLIENT_METADATA_USE_SESSION_DATABASE"
    CLIENT_PREFETCH_THREADS = "CLIENT_PREFETCH_THREADS"
    CLIENT_RESULT_CHUNK_SIZE = "CLIENT_RESULT_CHUNK_SIZE"
    CLIENT_RESULT_COLUMN_CASE_INSENSITIVE = "CLIENT_RESULT_COLUMN_CASE_INSENSITIVE"
    CLIENT_SESSION_CLONE = "CLIENT_SESSION_CLONE"
    CLIENT_SESSION_KEEP_ALIVE = "CLIENT_SESSION_KEEP_ALIVE"
    CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY = "CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY"
    CLIENT_TIMESTAMP_TYPE_MAPPING = "CLIENT_TIMESTAMP_TYPE_MAPPING"
    CSV_TIMESTAMP_FORMAT = "CSV_TIMESTAMP_FORMAT"
    C_API_QUERY_RESULT_FORMAT = "C_API_QUERY_RESULT_FORMAT"
    DATE_INPUT_FORMAT = "DATE_INPUT_FORMAT"
    DATE_OUTPUT_FORMAT = "DATE_OUTPUT_FORMAT"
    ENABLE_CONSOLE_OUTPUT = "ENABLE_CONSOLE_OUTPUT"
    ENABLE_UNLOAD_PHYSICAL_TYPE_OPTIMIZATION = "ENABLE_UNLOAD_PHYSICAL_TYPE_OPTIMIZATION"
    ERROR_ON_NONDETERMINISTIC_MERGE = "ERROR_ON_NONDETERMINISTIC_MERGE"
    ERROR_ON_NONDETERMINISTIC_UPDATE = "ERROR_ON_NONDETERMINISTIC_UPDATE"
    GEOGRAPHY_OUTPUT_FORMAT = "GEOGRAPHY_OUTPUT_FORMAT"
    GEOMETRY_OUTPUT_FORMAT = "GEOMETRY_OUTPUT_FORMAT"
    GO_QUERY_RESULT_FORMAT = "GO_QUERY_RESULT_FORMAT"
    JDBC_FORMAT_DATE_WITH_TIMEZONE = "JDBC_FORMAT_DATE_WITH_TIMEZONE"
    JDBC_QUERY_RESULT_FORMAT = "JDBC_QUERY_RESULT_FORMAT"
    JDBC_TREAT_DECIMAL_AS_INT = "JDBC_TREAT_DECIMAL_AS_INT"
    JDBC_TREAT_TIMESTAMP_NTZ_AS_UTC = "JDBC_TREAT_TIMESTAMP_NTZ_AS_UTC"
    JDBC_USE_SESSION_TIMEZONE = "JDBC_USE_SESSION_TIMEZONE"
    JSON_INDENT = "JSON_INDENT"
    JS_TREAT_INTEGER_AS_BIGINT = "JS_TREAT_INTEGER_AS_BIGINT"
    LANGUAGE = "LANGUAGE"
    LOCK_TIMEOUT = "LOCK_TIMEOUT"
    LOG_LEVEL = "LOG_LEVEL"
    MULTI_STATEMENT_COUNT = "MULTI_STATEMENT_COUNT"
    ODBC_QUERY_RESULT_FORMAT = "ODBC_QUERY_RESULT_FORMAT"
    ODBC_SCHEMA_CACHING = "ODBC_SCHEMA_CACHING"
    ODBC_USE_CUSTOM_SQL_DATA_TYPES = "ODBC_USE_CUSTOM_SQL_DATA_TYPES"
    PREVENT_UNLOAD_TO_INTERNAL_STAGES = "PREVENT_UNLOAD_TO_INTERNAL_STAGES"
    PYTHON_CONNECTOR_QUERY_RESULT_FORMAT = "PYTHON_CONNECTOR_QUERY_RESULT_FORMAT"
    PYTHON_SNOWPARK_USE_SCOPED_TEMP_OBJECTS = "PYTHON_SNOWPARK_USE_SCOPED_TEMP_OBJECTS"
    PYTHON_SNOWPARK_USE_SQL_SIMPLIFIER = "PYTHON_SNOWPARK_USE_SQL_SIMPLIFIER"
    QA_TEST_NAME = "QA_TEST_NAME"
    QUERY_RESULT_FORMAT = "QUERY_RESULT_FORMAT"
    QUERY_TAG = "QUERY_TAG"
    QUOTED_IDENTIFIERS_IGNORE_CASE = "QUOTED_IDENTIFIERS_IGNORE_CASE"
    READ_LATEST_WRITES = "READ_LATEST_WRITES"
    ROWS_PER_RESULTSET = "ROWS_PER_RESULTSET"
    S3_STAGE_VPCE_DNS_NAME = "S3_STAGE_VPCE_DNS_NAME"
    SEARCH_PATH = "SEARCH_PATH"
    SHOW_EXTERNAL_TABLE_KIND_AS_TABLE = "SHOW_EXTERNAL_TABLE_KIND_AS_TABLE"
    SIMULATED_DATA_SHARING_CONSUMER = "SIMULATED_DATA_SHARING_CONSUMER"
    SNOWPARK_HIDE_INTERNAL_ALIAS = "SNOWPARK_HIDE_INTERNAL_ALIAS"
    SNOWPARK_LAZY_ANALYSIS = "SNOWPARK_LAZY_ANALYSIS"
    SNOWPARK_REQUEST_TIMEOUT_IN_SECONDS = "SNOWPARK_REQUEST_TIMEOUT_IN_SECONDS"
    SNOWPARK_STORED_PROC_IS_FINAL_TABLE_QUERY = "SNOWPARK_STORED_PROC_IS_FINAL_TABLE_QUERY"
    SNOWPARK_USE_SCOPED_TEMP_OBJECTS = "SNOWPARK_USE_SCOPED_TEMP_OBJECTS"
    SQL_API_NULLABLE_IN_RESULT_SET = "SQL_API_NULLABLE_IN_RESULT_SET"
    SQL_API_QUERY_RESULT_FORMAT = "SQL_API_QUERY_RESULT_FORMAT"
    STATEMENT_QUEUED_TIMEOUT_IN_SECONDS = "STATEMENT_QUEUED_TIMEOUT_IN_SECONDS"
    STATEMENT_TIMEOUT_IN_SECONDS = "STATEMENT_TIMEOUT_IN_SECONDS"
    STRICT_JSON_OUTPUT = "STRICT_JSON_OUTPUT"
    TIMESTAMP_DAY_IS_ALWAYS_24H = "TIMESTAMP_DAY_IS_ALWAYS_24H"
    TIMESTAMP_INPUT_FORMAT = "TIMESTAMP_INPUT_FORMAT"
    TIMESTAMP_LTZ_OUTPUT_FORMAT = "TIMESTAMP_LTZ_OUTPUT_FORMAT"
    TIMESTAMP_NTZ_OUTPUT_FORMAT = "TIMESTAMP_NTZ_OUTPUT_FORMAT"
    TIMESTAMP_OUTPUT_FORMAT = "TIMESTAMP_OUTPUT_FORMAT"
    TIMESTAMP_TYPE_MAPPING = "TIMESTAMP_TYPE_MAPPING"
    TIMESTAMP_TZ_OUTPUT_FORMAT = "TIMESTAMP_TZ_OUTPUT_FORMAT"
    TIMEZONE = "TIMEZONE"
    TIME_INPUT_FORMAT = "TIME_INPUT_FORMAT"
    TIME_OUTPUT_FORMAT = "TIME_OUTPUT_FORMAT"
    TRACE_LEVEL = "TRACE_LEVEL"
    TRANSACTION_ABORT_ON_ERROR = "TRANSACTION_ABORT_ON_ERROR"
    TRANSACTION_DEFAULT_ISOLATION_LEVEL = "TRANSACTION_DEFAULT_ISOLATION_LEVEL"
    TWO_DIGIT_CENTURY_START = "TWO_DIGIT_CENTURY_START"
    UI_QUERY_RESULT_FORMAT = "UI_QUERY_RESULT_FORMAT"
    UNSUPPORTED_DDL_ACTION = "UNSUPPORTED_DDL_ACTION"
    USE_CACHED_RESULT = "USE_CACHED_RESULT"
    WEEK_OF_YEAR_POLICY = "WEEK_OF_YEAR_POLICY"
    WEEK_START = "WEEK_START"


class TaskState(ParseableEnum):
    """
    Represents the scheduling state of a task.
    https://docs.snowflake.com/en/sql-reference/sql/show-tasks#output
    Snowflake provides these as lowercase but we'll uppercase them.
    """

    STARTED = "STARTED"
    SUSPENDED = "SUSPENDED"


class WarehouseSize(ParseableEnum):
    """
    Represents the size options for a warehouse.

    Available sizes:
    - XSMALL
    - SMALL
    - MEDIUM
    - LARGE
    - XLARGE
    - XXLARGE
    - XXXLARGE
    - X4LARGE
    - X5LARGE (AWS-only)
    - X6LARGE (AWS-only)
    """

    XSMALL = "XSMALL"
    SMALL = "SMALL"
    MEDIUM = "MEDIUM"
    LARGE = "LARGE"
    XLARGE = "XLARGE"
    XXLARGE = "XXLARGE"
    XXXLARGE = "XXXLARGE"
    X4LARGE = "X4LARGE"
    X5LARGE = "X5LARGE"
    X6LARGE = "X6LARGE"

    @classmethod
    def synonyms(cls):
        return {
            "X-SMALL": "XSMALL",
            "X-LARGE": "XLARGE",
            "X2LARGE": "XXLARGE",
            "2X-LARGE": "XXLARGE",
            "X3LARGE": "XXXLARGE",
            "3X-LARGE": "XXXLARGE",
            "4X-LARGE": "X4LARGE",
            "5X-LARGE": "X5LARGE",
            "6X-LARGE": "X6LARGE",
        }


class FileType(ParseableEnum):
    CSV = "CSV"
    JSON = "JSON"
    AVRO = "AVRO"
    ORC = "ORC"
    PARQUET = "PARQUET"
    XML = "XML"


class Compression(ParseableEnum):
    AUTO = "AUTO"
    GZIP = "GZIP"
    BZ2 = "BZ2"
    BROTLI = "BROTLI"
    ZSTD = "ZSTD"
    DEFLATE = "DEFLATE"
    RAW_DEFLATE = "RAW_DEFLATE"
    LZO = "LZO"
    SNAPPY = "SNAPPY"
    NONE = "NONE"


class BinaryFormat(ParseableEnum):
    HEX = "HEX"
    BASE64 = "BASE64"
    UTF8 = "UTF8"


def resource_type_is_grant(resource_type: ResourceType) -> bool:
    return resource_type in (
        ResourceType.GRANT,
        ResourceType.GRANT_ON_ALL,
        ResourceType.ROLE_GRANT,
        ResourceType.GRANT_ON_ALL,
        ResourceType.FUTURE_GRANT,
    )



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ResourceType
from ..resource_name import ResourceName
from ..scope import AccountScope

from ..props import (
    Props,
    StringProp,
)


@dataclass(unsafe_hash=True)
class _Share(ResourceSpec):
    name: ResourceName
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class Share(NamedResource, Resource):
    """
    Description:
        Represents a share resource in Snowflake, which allows sharing data across Snowflake accounts.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-share

    Fields:
        name (string, required): The name of the share.
        owner (string or Role): The owner of the share. Defaults to "ACCOUNTADMIN".
        comment (string): A comment about the share.

    Python:

        ```python
        share = Share(
            name="some_share",
            comment="This is a snowflake share."
        )
        ```

    Yaml:

        ```yaml
        shares:
          - name: some_share
            comment: This is a snowflake share.
        ```
    """

    resource_type = ResourceType.SHARE
    props = Props(
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _Share

    def __init__(
        self,
        name: str,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data = _Share(
            name=self._name,
            owner=owner,
            comment=comment,
        )



from titan.blueprint import Blueprint
from titan.gitops import collect_resources_from_config
from titan.operations.connector import connect


def blueprint_plan(config, run_mode):
    session = connect()
    resources = collect_resources_from_config(config)
    blueprint = Blueprint(
        resources=resources,
        run_mode=run_mode,
    )
    plan_obj = blueprint.plan(session)
    return plan_obj


def blueprint_apply(plan, run_mode, dry_run):
    session = connect()
    blueprint = Blueprint(
        run_mode=run_mode,
        dry_run=dry_run,
    )
    blueprint.apply(session, plan)



from dataclasses import dataclass, field

from ..enums import ResourceType
from ..props import (
    BoolProp,
    FlagProp,
    IdentifierListProp,
    IntProp,
    Props,
    StringProp,
    TagsProp,
)
from ..resource_name import ResourceName
from ..scope import SchemaScope
from .resource import NamedResource, Resource, ResourceSpec
from .role import Role
from .tag import TaggableResource


@dataclass(unsafe_hash=True)
class _EventTable(ResourceSpec):
    name: ResourceName
    cluster_by: list[str] = None
    data_retention_time_in_days: int = None
    max_data_extension_time_in_days: int = None
    change_tracking: bool = False
    default_ddl_collation: str = None
    copy_grants: bool = field(default_factory=None, metadata={"fetchable": False})
    comment: str = None
    # row_access_policy: str = None
    owner: Role = "SYSADMIN"


class EventTable(NamedResource, TaggableResource, Resource):
    """
    Description:
        An event table captures events, including logged messages from functions and procedures.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-event-table

    Fields:
        name (string, required): The name of the event table.
        cluster_by (list): The expressions to cluster data by.
        data_retention_time_in_days (int): The number of days to retain data.
        max_data_extension_time_in_days (int): The maximum number of days to extend data retention.
        change_tracking (bool): Specifies whether change tracking is enabled. Defaults to False.
        default_ddl_collation (string): The default collation for DDL operations.
        copy_grants (bool): Specifies whether to copy grants. Defaults to False.
        comment (string): A comment for the event table.
        tags (dict): Tags associated with the event table.

    Python:

        ```python
        event_table = EventTable(
            name="some_event_table",
            cluster_by=["timestamp", "user_id"],
            data_retention_time_in_days=365,
            max_data_extension_time_in_days=30,
            change_tracking=True,
            default_ddl_collation="utf8",
            copy_grants=True,
            comment="This is a sample event table.",
            tags={"department": "analytics"}
        )
        ```

    Yaml:

        ```yaml
        event_tables:
          - name: some_event_table
            cluster_by:
              - timestamp
              - user_id
            data_retention_time_in_days: 365
            max_data_extension_time_in_days: 30
            change_tracking: true
            default_ddl_collation: utf8
            copy_grants: true
            comment: This is a sample event table.
            tags:
              department: analytics
        ```
    """

    resource_type = ResourceType.EVENT_TABLE
    props = Props(
        cluster_by=IdentifierListProp("cluster by", eq=False, parens=True),
        data_retention_time_in_days=IntProp("data_retention_time_in_days"),
        max_data_extension_time_in_days=IntProp("max_data_extension_time_in_days"),
        change_tracking=BoolProp("change_tracking"),
        default_ddl_collation=StringProp("default_ddl_collation"),
        copy_grants=FlagProp("copy grants"),
        comment=StringProp("comment", consume="with"),
        tags=TagsProp(),
    )
    scope = SchemaScope()
    spec = _EventTable

    def __init__(
        self,
        name: str,
        cluster_by: list[str] = None,
        data_retention_time_in_days: int = None,
        max_data_extension_time_in_days: int = None,
        change_tracking: bool = False,
        default_ddl_collation: str = None,
        copy_grants: bool = None,
        comment: str = None,
        owner: Role = "SYSADMIN",
        tags: dict[str, str] = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _EventTable = _EventTable(
            name=self._name,
            cluster_by=cluster_by,
            data_retention_time_in_days=data_retention_time_in_days,
            max_data_extension_time_in_days=max_data_extension_time_in_days,
            change_tracking=change_tracking,
            default_ddl_collation=default_ddl_collation,
            copy_grants=copy_grants,
            comment=comment,
            owner=owner,
        )
        self.set_tags(tags)



from setuptools import find_packages, setup


setup(
    name="titan-core",
    # Package version is managed by the string inside version.md. By default,
    # setuptools doesnt copy this file into the build package. So we direct
    # setuptools to include it using the `include_package_data=True` option
    # as well as the MANIFEST.in file which has the `include version.md` directive.
    version=open("version.md", encoding="utf-8").read().split(" ")[2],
    include_package_data=True,
    description="Titan Core: Snowflake infrastructure as code",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Titan-Systems/titan",
    author="TJ Murphy",
    packages=find_packages(include=["titan", "titan.*"]),
    python_requires=">=3.9",
    project_urls={
        "Homepage": "https://github.com/Titan-Systems/titan",
    },
    entry_points={
        "console_scripts": [
            "titan=titan.cli:titan_cli",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: SQL",
        "Topic :: Database",
    ],
    install_requires=[
        "click>=8.1.7",
        "inflection>=0.5.1",
        "pyparsing>=3.0.9",
        "pyyaml",
        "snowflake-connector-python>=3.7.0",
        "snowflake-snowpark-python>=1.14.0",
    ],
    extras_require={
        "dev": [
            "black",
            "build",
            "codespell==2.2.6",
            "pytest-cov",
            "pytest-profiling",
            "pytest-xdist",
            "pytest>=6.0",
            "python-dotenv",
            "ruff",
            "snowflake-cli-labs",
            "tabulate",
            "twine!=5.1.0",
        ]
    },
)



from dataclasses import dataclass, field

from ..enums import ResourceType
from ..props import (
    BoolProp,
    ColumnNamesProp,
    FlagProp,
    Props,
    QueryProp,
    StringProp,
    TagsProp,
)
from ..resource_name import ResourceName
from ..scope import SchemaScope
from .resource import NamedResource, Resource, ResourceSpec
from .role import Role
from .tag import TaggableResource


@dataclass(unsafe_hash=True)
class _View(ResourceSpec):
    name: ResourceName
    owner: Role = "SYSADMIN"
    secure: bool = False
    volatile: bool = None
    recursive: bool = None
    columns: list[dict] = None
    change_tracking: bool = False
    copy_grants: bool = field(default_factory=False, metadata={"fetchable": False})
    comment: str = None
    # TODO: remove this if parsing is feasible
    as_: str = field(default_factory=None, metadata={"fetchable": False})

    def __post_init__(self):
        super().__post_init__()
        if self.columns is not None and len(self.columns) == 0:
            raise ValueError("columns can't be empty")


class View(NamedResource, TaggableResource, Resource):
    """
    Description:
        Represents a view in Snowflake, which is a virtual table created by a stored query on the data.
        Views are used to simplify complex queries, improve security, or enhance performance.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-view

    Fields:
        name (string, required): The name of the view.
        owner (string or Role): The owner role of the view. Defaults to "SYSADMIN".
        secure (bool): Specifies if the view is secure.
        volatile (bool): Specifies if the view is volatile.
        recursive (bool): Specifies if the view is recursive.
        columns (list): A list of dictionaries specifying column details.
        tags (dict): A dictionary of tags associated with the view.
        change_tracking (bool): Specifies if change tracking is enabled.
        copy_grants (bool): Specifies if grants should be copied from the base table.
        comment (string): A comment for the view.
        as_ (string): The SELECT statement defining the view.

    Python:

        ```python
        view = View(
            name="some_view",
            owner="SYSADMIN",
            secure=True,
            as_="SELECT * FROM some_table"
        )
        ```

    Yaml:

        ```yaml
        views:
          - name: some_view
            owner: SYSADMIN
            secure: true
            as_: "SELECT * FROM some_table"
        ```
    """

    resource_type = ResourceType.VIEW
    props = Props(
        columns=ColumnNamesProp(),
        secure=FlagProp("secure"),
        volatile=FlagProp("volatile"),
        recursive=FlagProp("recursive"),
        tags=TagsProp(),
        change_tracking=BoolProp("change_tracking"),
        copy_grants=FlagProp("copy grants"),
        comment=StringProp("comment"),
        as_=QueryProp("as"),
    )
    scope = SchemaScope()
    spec = _View

    def __init__(
        self,
        name: str,
        owner: str = "SYSADMIN",
        secure: bool = False,
        volatile: bool = None,
        recursive: bool = None,
        columns: list[dict] = None,
        tags: dict[str, str] = None,
        change_tracking: bool = False,
        copy_grants: bool = False,
        comment: str = None,
        as_: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _View = _View(
            name=self._name,
            owner=owner,
            secure=secure,
            volatile=volatile,
            recursive=recursive,
            columns=columns,
            change_tracking=change_tracking,
            copy_grants=copy_grants,
            comment=comment,
            as_=as_,
        )
        self.set_tags(tags)



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ParseableEnum, ResourceType
from ..resource_name import ResourceName
from ..scope import SchemaScope
from ..props import (
    EnumProp,
    Props,
    StringProp,
    StringListProp,
)


class NetworkIdentifierType(ParseableEnum):
    IPV4 = "IPV4"
    AWSVPCEID = "AWSVPCEID"
    AZURELINKID = "AZURELINKID"
    HOST_PORT = "HOST_PORT"


class NetworkRuleMode(ParseableEnum):
    INGRESS = "INGRESS"
    INTERNAL_STAGE = "INTERNAL_STAGE"
    EGRESS = "EGRESS"


@dataclass(unsafe_hash=True)
class _NetworkRule(ResourceSpec):
    name: ResourceName
    type: NetworkIdentifierType
    value_list: list[str] = None
    mode: NetworkRuleMode = NetworkRuleMode.INGRESS
    comment: str = None
    owner: Role = "SYSADMIN"

    def __post_init__(self):
        super().__post_init__()
        if self.type == NetworkIdentifierType.HOST_PORT and self.mode != NetworkRuleMode.EGRESS:
            raise ValueError("When TYPE is HOST_PORT, MODE must be set to EGRESS.")

        if self.value_list is not None and len(self.value_list) == 0:
            raise ValueError("value_list must have at least one entry")


class NetworkRule(NamedResource, Resource):
    """
    Description:
        A Network Rule in Snowflake defines a set of network addresses, such as IP addresses or hostnames,
        that can be allowed or denied access to a Snowflake account. This helps in managing network traffic
        and securing access based on network policies.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-network-rule

    Fields:
        name (string, required): The name of the network rule.
        type (string or NetworkIdentifierType, required): The type of network identifier. Defaults to IPV4.
        value_list (list): A list of values associated with the network rule.
        mode (string or NetworkRuleMode): The mode of the network rule. Defaults to INGRESS.
        comment (string): A comment about the network rule.
        owner (string or Role): The owner role of the network rule. Defaults to "SYSADMIN".

    Python:

        ```python
        network_rule = NetworkRule(
            name="some_network_rule",
            type="IPV4",
            value_list=["192.168.1.1", "192.168.1.2"],
            mode="INGRESS",
            comment="Example network rule"
        )
        ```

    Yaml:

        ```yaml
        network_rules:
          - name: some_network_rule
            type: IPV4
            value_list: ["192.168.1.1", "192.168.1.2"]
            mode: INGRESS
            comment: "Example network rule"
        ```
    """

    resource_type = ResourceType.NETWORK_RULE
    props = Props(
        type=EnumProp("type", NetworkIdentifierType),
        value_list=StringListProp("value_list", parens=True, eq=True),
        mode=EnumProp("mode", NetworkRuleMode),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _NetworkRule

    def __init__(
        self,
        name: str,
        type: NetworkIdentifierType = NetworkIdentifierType.IPV4,
        value_list: list[str] = None,
        mode: NetworkRuleMode = NetworkRuleMode.INGRESS,
        comment: str = None,
        owner: str = "SYSADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _NetworkRule = _NetworkRule(
            name=self._name,
            type=type,
            value_list=value_list,
            mode=mode,
            comment=comment,
            owner=owner,
        )



from dataclasses import dataclass, field

from ..enums import ResourceType
from ..props import (
    IdentifierProp,
    Props,
    StringProp,
    TagsProp,
)
from ..resource_name import ResourceName
from ..scope import SchemaScope
from .resource import NamedResource, Resource, ResourceSpec
from .role import Role
from .tag import TaggableResource
from .warehouse import Warehouse

# TODO: I can't get version to work at all with Snowflake, I suspect it's buggy.


@dataclass(unsafe_hash=True)
class _Notebook(ResourceSpec):
    name: ResourceName
    # version: str = None
    from_: str = field(default_factory=None, metadata={"fetchable": False})
    main_file: str = None
    comment: str = None
    default_version: str = None
    query_warehouse: Warehouse = None
    owner: Role = "SYSADMIN"


class Notebook(NamedResource, Resource):
    resource_type = ResourceType.NOTEBOOK
    props = Props(
        # version=StringProp("version", eq=False),
        from_=StringProp("from", eq=False),
        main_file=StringProp("main_file"),
        comment=StringProp("comment"),
        default_version=StringProp("default_version"),
        query_warehouse=IdentifierProp("query_warehouse"),
        tags=TagsProp(),
    )
    scope = SchemaScope()
    spec = _Notebook

    def __init__(
        self,
        name: str,
        # version: str = None,
        from_: str = None,
        main_file: str = None,
        comment: str = None,
        default_version: str = None,
        query_warehouse: str = None,
        owner: Role = "SYSADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _Notebook = _Notebook(
            name=self._name,
            # version=version,
            from_=from_,
            main_file=main_file,
            comment=comment,
            default_version=default_version,
            query_warehouse=query_warehouse,
            owner=owner,
        )



import pytest

from tests.helpers import get_json_fixtures
from titan import resources as res

JSON_FIXTURES = list(get_json_fixtures())


@pytest.fixture(
    params=JSON_FIXTURES,
    ids=[resource_cls.__name__ for resource_cls, _ in JSON_FIXTURES],
    scope="function",
)
def json_fixture(request):
    resource_cls, data = request.param
    yield resource_cls, data


def test_resource_requires(json_fixture):
    role = res.Role(name="dummy")
    resource_cls, data = json_fixture
    resource = resource_cls(**data)
    assert role.requires(resource) is None
    assert resource in role.refs



import pytest

from titan import resources as res
from titan.enums import ResourceType
from titan.resources.resource import ResourceHasContainerException, WrongContainerException


def test_account_can_add_database():
    account = res.Account("SOME_ACCOUNT", "SOME_LOCATOR")
    database = res.Database("SOME_DATABASE")
    account.add(database)
    assert database in account.items(resource_type=ResourceType.DATABASE)


def test_account_cannot_add_schema():
    account = res.Account("SOME_ACCOUNT", "SOME_LOCATOR")
    schema = res.Schema("SOME_SCHEMA")
    with pytest.raises(WrongContainerException):
        account.add(schema)


def test_database_can_add_schema():
    database = res.Database("SOME_DATABASE")
    schema = res.Schema("SOME_SCHEMA")
    database.add(schema)
    assert schema in database.items(resource_type=ResourceType.SCHEMA)


def test_database_cannot_add_user():
    database = res.Database("SOME_DATABASE")
    user = res.User("SOME_USER")
    with pytest.raises(WrongContainerException):
        database.add(user)


def test_database_cannot_add_view():
    database = res.Database("SOME_DATABASE")
    view = res.View("SOME_VIEW")
    with pytest.raises(WrongContainerException):
        database.add(view)


def test_schema_can_add_view():
    schema = res.Schema("SOME_SCHEMA")
    view = res.View("SOME_VIEW")
    schema.add(view)
    assert view in schema.items(resource_type=ResourceType.VIEW)


def test_schema_cannot_add_user():
    schema = res.Schema("SOME_SCHEMA")
    user = res.User("SOME_USER")
    with pytest.raises(WrongContainerException):
        schema.add(user)


def test_remove_resource():
    database = res.Database("SOME_DATABASE")
    schema = res.Schema("SOME_SCHEMA")
    database.add(schema)
    assert schema in database.items(resource_type=ResourceType.SCHEMA)
    database.remove(schema)
    assert schema not in database.items(resource_type=ResourceType.SCHEMA)


def test_resource_already_belongs_to_container():
    schema1 = res.Schema("SOME_SCHEMA1")
    schema2 = res.Schema("SOME_SCHEMA2")
    view = res.View("SOME_VIEW")

    schema1.add(view)
    assert view in schema1.items(resource_type=ResourceType.VIEW)

    with pytest.raises(ResourceHasContainerException):
        schema2.add(view)



import pytest

from titan import resources as res
from titan.enums import ResourceType
from titan.resources.resource import ResourcePointer


def test_from_sql_fqn_parsing():
    grant = res.Grant.from_sql('GRANT USAGE ON SCHEMA "My_databasE".my_schema TO ROLE my_role')
    assert isinstance(grant.to, ResourcePointer)
    assert grant.to.resource_type == ResourceType.ROLE
    assert grant.to.name == "my_role"
    assert grant.on == '"My_databasE".MY_SCHEMA'
    schema_ref = None
    for ref in grant.refs:
        if ref.resource_type == ResourceType.SCHEMA and ref.name == "my_schema":
            schema_ref = ref
    assert schema_ref
    assert schema_ref.container.name == '"My_databasE"'
